#!/usr/bin/env python3

# To test basic functionality run python3 -m doctest [-v] sorts.py

import utils

def find_key(array, key):
    """ Return index of a key if present, raise IndexError otherwise

        >>> find_key([1, 2, 3], 2)
        1
        >>> find_key([2, 3], 4)
        Traceback (most recent call last):
        ...
        IndexError
    """

    raise IndexError


def is_sorted(array):
    """ Check whether the array is sorted

         >>> is_sorted([])
         True
         >>> is_sorted([1])
         True
         >>> is_sorted([1, 2, 2, 4, 5])
         True
         >>> is_sorted([6, 5, 3, 4])
         False
    """

    return True


def bubble_sort(array):
    """ Sort array using Bubble sort algorithm

        >>> bubble_sort([])
        []
        >>> bubble_sort([1, 1, 1])
        [1, 1, 1]
        >>> bubble_sort([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]
    """

    # Compare adjacent elements, swap them if necessary
    # inner loop -> moves the biggest element to the end of the array
    # outer loop -> controls the maximum number of iterations needed to sort an array

    return array


def bubble_sort_adaptive(array):
    """ Sort array using adaptive form of Bubble sort algorithm

        >>> bubble_sort_adaptive([])
        []
        >>> bubble_sort_adaptive([1, 1, 1])
        [1, 1, 1]
        >>> bubble_sort_adaptive([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]
    """

    # End the procedure when the array is sorted
    
    return array


def selection_sort(array):
    """ Sort array using Selection sort algorithm

        >>> selection_sort([])
        []
        >>> selection_sort([1, 1, 1])
        [1, 1, 1]
        >>> selection_sort([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]
    """

    # Move the smallest element to the beginning
    # inner loop -> find the index of the smallest element
    # outer loop -> swap the k-th smallest element with element on the k-th position

    return array


def quick_sort(array):
    """ Sort array using Quick sort algorithm

        >>> quick_sort([])
        []
        >>> quick_sort([1, 1, 1])
        [1, 1, 1]
        >>> quick_sort([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]
    """

    # Arrays of size <= 1 are sorted
    # Choose a pivot element
    # Create two lists (< pivot; >= pivot)
    # Sort them recursively using Quick sort, join the results appropriately

    return array


def merge(a, b):
    """ Merge two sorted arrays into one 

        >>> merge([], [1, 2])
        [1, 2]
        >>> merge([], [])
        []
        >>> merge([1, 3, 5, 7], [2, 4, 6])
        [1, 2, 3, 4, 5, 6, 7]

    """

    # If one of the arrays is empty, return the second one
    # Compare first elements of the arrays, return the smaller one 
    # joined with the result of the merge function without this element

    return a + b


def merge_iterative(a, b):
    """ Merge two sorted arrays into one, avoid recursion

        >>> merge_iterative([], [1, 2])
        [1, 2]
        >>> merge_iterative([], [])
        []
        >>> merge_iterative([1, 3, 5, 7], [2, 4, 6])
        [1, 2, 3, 4, 5, 6, 7]

    """

    # Same as merge, but use loop instead of the recursion

    return a + b 


def merge_sort(array):
    """ Sort array using Merge sort algorithm

        >>> merge_sort([])
        []
        >>> merge_sort([1, 1, 1])
        [1, 1, 1]
        >>> merge_sort([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]

    """

    # Arays of size <= 1 are sorted
    # Compute the middle index to the array
    # Sort the first and second half of the array with Merge sort
    # Merge sorted halves into the final sorted array

    return array

if __name__ == "__main__":

    array = utils.generate_array(20, array_type="random")

    print(array)
    selection_sort(array)
    print(array)

    # utils.print_times(selection_sort, 100, 5, array_type="random", runs=3)