#!/usr/bin/env python3

import utils

def find_key(array, key):
    """ Return index of a key if present, raise IndexError otherwise

        >>> find_key([1, 2, 3], 2)
        1
        >>> find_key([2, 3], 4)
        Traceback (most recent call last):
        ...
        IndexError
    """

    for index, item in enumerate(array):
        if item == key:
            return index

    raise IndexError


def is_sorted(array):
    """ Check whether the array is sorted

         >>> is_sorted([])
         True
         >>> is_sorted([1])
         True
         >>> is_sorted([1, 2, 2, 4, 5])
         True
         >>> is_sorted([6, 5, 3, 4])
         False
    """

    for i in range(len(array) - 1):
        if array[i] > array[i + 1]:
            return False

    return True


def bubble_sort(array):
    """ Sort array using Bubble sort algorithm

        >>> bubble_sort([])
        []
        >>> bubble_sort([1, 1, 1])
        [1, 1, 1]
        >>> bubble_sort([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]
    """

    for i in range(len(array)):
        for j in range(len(array) - i - 1):
            if array[j] > array[j + 1]:
                array[j], array[j + 1] = array[j + 1], array[j]

    return array


def bubble_sort_adaptive(array):
    """ Sort array using adaptive form of Bubble sort algorithm

        >>> bubble_sort_adaptive([])
        []
        >>> bubble_sort_adaptive([1, 1, 1])
        [1, 1, 1]
        >>> bubble_sort_adaptive([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]
    """

    for i in range(len(array)):
        swapped = False
        for j in range(len(array) - i - 1):
            if array[j] > array[j + 1]:
                array[j], array[j + 1] = array[j + 1], array[j]
                swapped = True

        if not swapped:
            break

    return array


def selection_sort(array):
    """ Sort array using Selection sort algorithm

        >>> selection_sort([])
        []
        >>> selection_sort([1, 1, 1])
        [1, 1, 1]
        >>> selection_sort([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]
    """

    for i in range(len(array)):
        min_idx = i
        for j in range(i, len(array)):
            if array[j] < array[min_idx]:
                min_idx = j

        array[i], array[min_idx] = array[min_idx], array[i]

    return array


def quick_sort(array):
    """ Sort array using Quick sort algorithm

        >>> quick_sort([])
        []
        >>> quick_sort([1, 1, 1])
        [1, 1, 1]
        >>> quick_sort([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]
    """

    if len(array) <= 1:
        return array

    pivot = array[0]
    smaller = [x for x in array[1 :] if x < pivot]
    bigger = [x for x in array[1 :] if x >= pivot]

    return quick_sort(smaller) + [pivot] + quick_sort(bigger)


def merge(a, b):
    """ Merge two sorted arrays into one 

        >>> merge([], [1, 2])
        [1, 2]
        >>> merge([], [])
        []
        >>> merge([1, 3, 5, 7], [2, 4, 6])
        [1, 2, 3, 4, 5, 6, 7]

    """

    if not len(a):
        return b
    if not len(b):
        return a

    if a[0] < b[0]:
        return [a[0]] + merge(a[1:], b)
    else:
        return [b[0]] + merge(a, b[1:])


def merge_iterative(a, b):
    """ Merge two sorted arrays into one, avoid recursion

        >>> merge_iterative([], [1, 2])
        [1, 2]
        >>> merge_iterative([], [])
        []
        >>> merge_iterative([1, 3, 5, 7], [2, 4, 6])
        [1, 2, 3, 4, 5, 6, 7]

    """

    i = 0
    j = 0

    output = []

    while i < len(a) and j < len(b):
        if a[i] < b[j]:
            output.append(a[i])
            i += 1
        else:
            output.append(b[j])
            j += 1

    if i == len(a):
        output.extend(b[j:])
    if j == len(b):
        output.extend(a[i:])

    return output


def merge_sort(array):
    """ Sort array using Merge sort algorithm

        >>> merge_sort([])
        []
        >>> merge_sort([1, 1, 1])
        [1, 1, 1]
        >>> merge_sort([7, 4, 1, 3, 8])
        [1, 3, 4, 7, 8]

    """

    if len(array) <= 1:
        return array

    mid = len(array) // 2

    return merge_iterative(merge_sort(array[:mid]), merge_sort(array[mid:]))


if __name__ == "__main__":

    utils.print_times(selection_sort, 1000, 5, array_type="random", runs=3)